﻿using System;
using System.Text;
using System.Windows.Forms;

namespace RPGSoundController
{
    public partial class MainForm : Form, INetworkClient
    {
        public MainForm()
        {
            network = new Network(this);
            player = new Player();
            InitializeComponent();
            if (!Options.Default.ShowMesssages)
            {
                Options.Default.ShowMesssages = true;
                SwitchMessages();
            }
            Messages.Instance.MessageReceived += new MessageReceivedHandler(Instance_MessageReceived);
            network.InitConnectionData();
            if (Options.Default.PlayerPath == "RPG-SoundMixer")
            {
                ShowOptions();
            }
            else
            {
                ConnectPlayer();
                UpdatePlayerData();
            }
            network.StartUdpBroadcast();
            broadCastTimer.Tick += new EventHandler(broadCastTimer_Tick);
            broadCastTimer.Enabled = true;
            network.ListenForClient();
        }

        private void Instance_MessageReceived(Message m)
        {
            if (InvokeRequired)
            {
                Invoke(new MessageReceivedHandler(HandleMessage), m);
            }
            else
            {
                HandleMessage(m);
            }
        }

        private void HandleMessage(Message m)
        {
            if (messagesBox != null)
                messagesBox.Items.Add(m.Type + ": " + m.Text);
        }

        private void Shutdown()
        {
            if (network.ClientConnected)
            {
                network.DisconnectClient(false);
            }
            else
            {
                network.StopListenForClient();
                System.Threading.Thread.Sleep(400);
                if (network.ClientConnected)
                {
                    network.DisconnectClient(false);
                }
            }
            broadCastTimer.Enabled = false;
            network.StopUdpBroadcast();
            player.CloseConnection();
            Options.Default.Save();
        }

        private bool warnOnNetworkFail = true;

        void broadCastTimer_Tick(object sender, EventArgs e)
        {
            if (!network.SendUdpPacket())
            {
                if (warnOnNetworkFail)
                {
                    warnOnNetworkFail = false;
                    MessageBox.Show(this, Resources.NoStatusInfoError, "RPG SoundController", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private delegate void SendKeyInvoker(Keys key);

        private KeysConverter converter = new KeysConverter();

        private void HandleKey(Keys key)
        {
            player.SendKeyToPlayer(key);
            String text = converter.ConvertToInvariantString(key);
            if (keyLabel.Text != text)
            {
                keyLabel.Text = text;
                deleteKeyVisualsTimer.Start();
            }
        }

        public void KeyReceived(Keys key)
        {
            this.Invoke(new SendKeyInvoker(HandleKey), new object[] { key });
        }

        public void ClientDataChanged()
        {
            this.Invoke(new MethodInvoker(UpdateClientData));
        }

        private void UpdateClientData()
        {
            if (network.ClientConnected)
            {
                clientStateLabel.Text = Resources.ConnectedWith + network.ClientName;
                clientStateLabel.ForeColor = System.Drawing.Color.DarkGreen;
                disconnectButton.Enabled = true;
            }
            else
            {
                clientStateLabel.Text = Resources.NotConnected;
                clientStateLabel.ForeColor = System.Drawing.Color.Red;
                disconnectButton.Enabled = false;
            }
        }

        private void UpdatePlayerData()
        {
            if (player.Connected)
            {
                playerStatusLabel.ForeColor = System.Drawing.Color.DarkGreen;
                playerStatusLabel.Text = Resources.Connected;
            }
            else
            {
                playerStatusLabel.ForeColor = System.Drawing.Color.Red;
                playerStatusLabel.Text = Resources.NotConnected;
            }
        }

        private void disconnectButton_Click(object sender, EventArgs e)
        {
            network.DisconnectClient(true);
        }

        private void searchPlayerButton_Click(object sender, EventArgs e)
        {
            player.CloseConnection();
            ConnectPlayer();
            UpdatePlayerData();
        }

        private void deleteKeyVisualsTimer_Tick(object sender, EventArgs e)
        {
            keyLabel.Text = String.Empty;
            deleteKeyVisualsTimer.Stop();
        }

        private void ConnectPlayer()
        {
            player.InitConnection();
        }

        private void optionsButton_Click(object sender, EventArgs e)
        {
            ShowOptions();
        }

        private void ShowOptions()
        {
            OptionsDialog dialog = new OptionsDialog();
            String oldPlayerPath = Options.Default.PlayerPath;
            DialogResult result = dialog.ShowDialog(this);
            if (result == DialogResult.OK)
            {
                if (network.ClientConnected)
                {
                    network.DisconnectClient(true);
                }
                if (oldPlayerPath != Options.Default.PlayerPath)
                {
                    player.CloseConnection();
                    ConnectPlayer();
                    UpdatePlayerData();
                }
            }
        }

        private Network network;

        private Player player;

        private void messagesButton_Click(object sender, EventArgs e)
        {
            SwitchMessages();
        }

        private void SwitchMessages()
        {
            const int heightDiff = 148;
            bool showMessages = !Options.Default.ShowMesssages;
            int diff = showMessages ? heightDiff : -heightDiff;
            this.Height += diff;
            messagesButton.Text = Resources.Messages + (showMessages ? "<<" : ">>");
            messagesBox.Visible = showMessages;
            Options.Default.ShowMesssages = showMessages;
        }

    }
}
