﻿using System;
using System.Text;
using System.Net.Sockets;
using System.Net;

namespace RPGSoundController
{
    interface INetworkClient
    {
        void KeyReceived(System.Windows.Forms.Keys key);
        void ClientDataChanged();
    }

    class Network
    {
        public Network(INetworkClient client)
        {
            networkClient = client;
            ClientConnected = false;
            ClientName = String.Empty;
        }

        private static int udpPacketCount = 0;

        public bool SendUdpPacket()
        {
            if (udpClient == null)
            {
                return true;
            }
            try
            {
                ++udpPacketCount;
                if (udpPacketCount == 1)
                {
                    Messages.AddMessage(MessageType.Debug, String.Format(Resources.UDPSending, udpString));
                }
                udpClient.Send(udpPacket, udpPacket.Length, new IPEndPoint(IPAddress.Parse("255.255.255.255"), Options.Default.UdpPort));
                return true;
            }
            catch (SocketException e)
            {
                Messages.AddMessage(MessageType.Warning, Resources.UDPError + e.Message);
                return false;
            }
        }

        public void StartUdpBroadcast()
        {
            Messages.AddMessage(MessageType.Info, Resources.StartBroadcast);
            udpClient = new UdpClient();
            udpClient.EnableBroadcast = true;
            udpPacketCount = 0;
        }

        public void StopUdpBroadcast()
        {
            Messages.AddMessage(MessageType.Info, Resources.StopBroadcast);
            udpClient.Close();
            udpClient = null;
        }

        public void ListenForClient()
        {
            lock (this)
            {
                continueListenForClients = true;
            }
            System.Threading.Thread listenThread = new System.Threading.Thread(new System.Threading.ThreadStart(ListenInThread));
            listenThread.Start();
        }

        public void StopListenForClient()
        {
            lock (this)
            {
                continueListenForClients = false;
            }
        }

        public void ListenInThread()
        {
            bool goOn = true;
            TcpListener listener = new TcpListener(tcpListenAddress, Options.Default.TcpListenPort);
            try
            {
                listener.Start(1);
                while (goOn)
                {
                    if (listener.Pending())
                    {
                        TcpClient client = listener.AcceptTcpClient();
                        if (InitConnection(client))
                        {
                            StopListenForClient();
                            break;
                        }
                    }
                    lock (this)
                    {
                        goOn = continueListenForClients;
                    }
                    if (goOn) System.Threading.Thread.Sleep(100);
                }
            }
            catch (Exception e)
            {
                Messages.AddMessage(MessageType.Error, Resources.ClientListenError + e.Message);
            }
            finally
            {
                listener.Stop();
            }
        }

        private static bool ReadFromStream(NetworkStream stream, Byte[] buffer, int size, int timeout)
        {
            int read = 0;
            int time = 0;
            while (read < size)
            {
                while (!stream.DataAvailable)
                {
                    if (time > timeout) return false;
                    time += 5;
                    System.Threading.Thread.Sleep(5);
                }
                buffer[read++] = (byte)stream.ReadByte();
            }
            return true;
        }

        private bool InitConnection(TcpClient aClient)
        {
            Byte[] buffer = new Byte[1000];
            NetworkStream stream = aClient.GetStream();
            bool success = ReadFromStream(stream, buffer, 4, 2000);
            if (!success) return false;
            int length = 0;
            String s = System.Text.Encoding.UTF8.GetString(buffer, 0, 4);
            Messages.AddMessage(MessageType.Debug, Resources.ClientLengthReceived + s);
            success = Int32.TryParse(s, out length);
            if (!success)
            {
                Messages.AddMessage(MessageType.Debug, Resources.WrongClientLength);
                return false;
            }
            if (length > 1000)
            {
                Messages.AddMessage(MessageType.Debug, Resources.LengthTooHigh);
                return false;
            }
            success = ReadFromStream(stream, buffer, length, 2000);
            if (!success)
            {
                Messages.AddMessage(MessageType.Debug, Resources.NoClientID);
                return false;
            }
            ClientName = System.Text.Encoding.UTF8.GetString(buffer, 0, length);
            lock (this)
            {
                client = aClient;
                ClientConnected = true;
            }
            Messages.AddMessage(MessageType.Info, String.Format(Resources.ClientConnected, ClientName));
            networkClient.ClientDataChanged();
            System.Threading.Thread commandThread = new System.Threading.Thread(ListenForKeys);
            commandThread.Start();
            StopUdpBroadcast();
            return true;
        }

        public void DisconnectClient(bool listenAgain)
        {
            lock (this)
            {
                continueListenForKeys = false;
                client.Client.Close();
                client.Close();
                client = null;
            }
            ClientConnected = false;
            Messages.AddMessage(MessageType.Info, Resources.ClientDisconnected);
            networkClient.ClientDataChanged();
            if (listenAgain)
            {
                StartUdpBroadcast();
                ListenForClient();
            }
        }

        private System.Windows.Forms.KeysConverter converter = new System.Windows.Forms.KeysConverter();

        private System.Windows.Forms.Keys MapKeyCodeToKey(Byte[] keyCode)
        {
            try
            {
                if (keyCode[0] == 0)
                {
                    String s = System.Text.Encoding.ASCII.GetString(keyCode, 1, 1);
                    return (System.Windows.Forms.Keys)converter.ConvertFromInvariantString(s);
                }
                else if (keyCode[0] == 1)
                {
                    switch (keyCode[1])
                    {
                        case 1: return System.Windows.Forms.Keys.F1;
                        case 2: return System.Windows.Forms.Keys.F2;
                        case 3: return System.Windows.Forms.Keys.F3;
                        case 4: return System.Windows.Forms.Keys.F4;
                        case 5: return System.Windows.Forms.Keys.F5;
                        case 6: return System.Windows.Forms.Keys.F6;
                        case 7: return System.Windows.Forms.Keys.F7;
                        case 8: return System.Windows.Forms.Keys.F8;
                        case 9: return System.Windows.Forms.Keys.F9;
                        case 10: return System.Windows.Forms.Keys.F10;
                        case 11: return System.Windows.Forms.Keys.F11;
                        case 12: return System.Windows.Forms.Keys.F12;
                    }
                }
                else if (keyCode[0] == 2)
                {
                    switch (keyCode[1])
                    {
                        case 0: return System.Windows.Forms.Keys.NumPad0;
                        case 1: return System.Windows.Forms.Keys.NumPad1;
                        case 2: return System.Windows.Forms.Keys.NumPad2;
                        case 3: return System.Windows.Forms.Keys.NumPad3;
                        case 4: return System.Windows.Forms.Keys.NumPad4;
                        case 5: return System.Windows.Forms.Keys.NumPad5;
                        case 6: return System.Windows.Forms.Keys.NumPad6;
                        case 7: return System.Windows.Forms.Keys.NumPad7;
                        case 8: return System.Windows.Forms.Keys.NumPad8;
                        case 9: return System.Windows.Forms.Keys.NumPad9;
                    }
                }
                else if (keyCode[0] == 3)
                {
                    switch (keyCode[1])
                    {
                        case 0: return System.Windows.Forms.Keys.Insert;
                        case 1: return System.Windows.Forms.Keys.Delete;
                        case 2: return System.Windows.Forms.Keys.Home;
                        case 3: return System.Windows.Forms.Keys.End;
                        case 4: return System.Windows.Forms.Keys.PageUp;
                        case 5: return System.Windows.Forms.Keys.PageDown;
                        case 6: return System.Windows.Forms.Keys.Left;
                        case 7: return System.Windows.Forms.Keys.Right;
                        case 8: return System.Windows.Forms.Keys.Up;
                        case 9: return System.Windows.Forms.Keys.Down;
                        case 10: return System.Windows.Forms.Keys.Space;
                        case 11: return System.Windows.Forms.Keys.Return;
                        case 12: return System.Windows.Forms.Keys.Escape;
                        case 13: return System.Windows.Forms.Keys.Tab;
                        case 14: return System.Windows.Forms.Keys.Back;
                        case 15: return System.Windows.Forms.Keys.OemPeriod;
                        case 16: return System.Windows.Forms.Keys.OemSemicolon;
                        case 17: return System.Windows.Forms.Keys.Oem4;
                        case 18: return System.Windows.Forms.Keys.Oemcomma;
                        case 19: return System.Windows.Forms.Keys.Oem2;
                        case 20: return System.Windows.Forms.Keys.OemMinus;
                        case 21: return System.Windows.Forms.Keys.Oem3;
                    }
                }
                return System.Windows.Forms.Keys.Escape;
            }
            catch (Exception e)
            {
                Messages.AddMessage(MessageType.Error, String.Format(Resources.InvalidKeyCode, e.Message));
                return System.Windows.Forms.Keys.Escape;
            }
        }

        private void ListenForKeys()
        {
            try
            {
                bool goOn = true;
                bool disconnect = false;
                while (goOn && client != null)
                {
                    int command = -1;
                    lock (this)
                    {
                        if (client != null && client.GetStream().DataAvailable)
                        {
                            client.GetStream().ReadTimeout = 100;
                            try
                            {
                                command = client.GetStream().ReadByte();
                            }
                            catch (System.IO.IOException)
                            {
                                command = -1;
                            }
                        }
                    }
                    if (command != -1)
                    {
                        Messages.AddMessage(MessageType.Debug, String.Format(Resources.CommandReceived, command));
                    }
                    if (command == 0)
                    {
                        Byte[] keyCode = new Byte[2];
                        bool success = false;
                        lock (this)
                        {
                            success = client != null && ReadFromStream(client.GetStream(), keyCode, 2, 500);
                        }
                        if (success)
                        {
                            Messages.AddMessage(MessageType.Debug, String.Format(Resources.KeyReceived, keyCode[0], keyCode[1]));
                            networkClient.KeyReceived(MapKeyCodeToKey(keyCode));
                        }
                    }
                    else if (command == 1)
                    {
                        Messages.AddMessage(MessageType.Debug, Resources.DisconnectingClient);
                        disconnect = true;
                        break;
                    }
                    else if (command == 5)
                    {
                        // this is ping
                        Messages.AddMessage(MessageType.Debug, Resources.PingReceived);
                    }
                    lock (this)
                    {
                        goOn = continueListenForKeys;
                    }
                    if (goOn)
                    {
                        System.Threading.Thread.Sleep(5);
                    }
                }
                if (disconnect)
                {
                    DisconnectClient(true);
                }
            }
            catch (Exception e)
            {
                Messages.AddMessage(MessageType.Error, Resources.KeyListenError + e.Message);
            }
        }

        public void InitConnectionData()
        {
            StringBuilder str = new StringBuilder();
            String machineName = Dns.GetHostName();
            str.Append(machineName);
            str.Append("|");
            str.Append(Options.Default.TcpListenPort);
            foreach (IPAddress address in Dns.GetHostAddresses(machineName))
            {
                if (IPAddress.IsLoopback(address)) continue;
                str.Append("|");
                str.Append(address.ToString());
                tcpListenAddress = address;
                break;
            }
            udpString = str.ToString();
            udpPacket = System.Text.Encoding.UTF8.GetBytes(udpString);
        }

        private Byte[] udpPacket;

        private String udpString;

        private IPAddress tcpListenAddress = null;

        private bool continueListenForClients = true;

        private bool continueListenForKeys = true;

        public bool ClientConnected { get; set; }

        public String ClientName { get; set; }

        private TcpClient client = null;

        private UdpClient udpClient = null;

        private INetworkClient networkClient;
   }
}
